@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2011-2012 Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2011-2013 Red Hat, Inc. All rights reserved.
  *
  * This file is part of LVM2.
  *
@@ -9,25 +9,22 @@
  *
  * You should have received a copy of the GNU Lesser General Public License
  * along with this program; if not, write to the Free Software Foundation,
- * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
-#include "lib.h"
-#include "toolcontext.h"
-#include "metadata.h"
-#include "segtype.h"
-#include "text_export.h"
-#include "config.h"
-#include "activate.h"
-#include "str_list.h"
-#include "defaults.h"
+#include "base/memory/zalloc.h"
+#include "lib/misc/lib.h"
+#include "lib/display/display.h"
+#include "lib/metadata/metadata.h"
+#include "lib/metadata/segtype.h"
+#include "lib/format_text/text_export.h"
+#include "lib/config/config.h"
+#include "lib/activate/activate.h"
+#include "lib/datastruct/str_list.h"
 
-#ifdef DMEVENTD
-#  include "libdevmapper-event.h"
-#endif
-
-/* Dm kernel module name for thin provisiong */
-#define THIN_MODULE "thin-pool"
+/* Dm kernel module name for thin provisioning */
+static const char _thin_pool_module[] = "thin-pool";
+static const char _thin_module[] = "thin";
 
 /*
  * Macro used as return argument - returns 0.
@@ -37,18 +34,27 @@
 	log_error(t " segment %s of logical volume %s.", ## p, \
 		  dm_config_parent_name(sn), seg->lv->name), 0;
 
-static int _thin_target_present(struct cmd_context *cmd,
-				const struct lv_segment *seg,
-				unsigned *attributes);
+/* TODO: using static field here, maybe should be a part of segment_type */
+static unsigned _feature_mask;
 
-static const char *_thin_pool_name(const struct lv_segment *seg)
+static void _thin_pool_display(const struct lv_segment *seg)
 {
-	return seg->segtype->name;
+	log_print("  Chunk size\t\t%s",
+		  display_size(seg->lv->vg->cmd, seg->chunk_size));
+	log_print("  Discards\t\t%s", get_pool_discards_name(seg->discards));
+	log_print("  Thin count\t\t%u",
+		  dm_list_size(&seg->lv->segs_using_this_lv));
+	log_print("  Transaction ID\t%" PRIu64, seg->transaction_id);
+	log_print("  Zero new blocks\t%s",
+		  (seg->zero_new_blocks == THIN_ZERO_YES) ? "yes" : "no");
+
+	log_print(" ");
 }
 
 static int _thin_pool_add_message(struct lv_segment *seg,
 				  const char *key,
-				  const struct dm_config_node *sn)
+				  const struct dm_config_node *sn,
+				  struct dm_hash_table *lv_hash)
 {
 	const char *lv_name = NULL;
 	struct logical_volume *lv = NULL;
@@ -57,23 +63,17 @@ static int _thin_pool_add_message(struct lv_segment *seg,
 
 	/* Message must have only one from: create, delete */
 	if (dm_config_get_str(sn, "create", &lv_name)) {
-		if (!(lv = find_lv(seg->lv->vg, lv_name)))
+		if (!(lv = dm_hash_lookup(lv_hash, lv_name)))
 			return SEG_LOG_ERROR("Unknown LV %s for create message in",
 					     lv_name);
 		/* FIXME: switch to _SNAP later, if the created LV has an origin */
 		type = DM_THIN_MESSAGE_CREATE_THIN;
-	}
-
-	if (!dm_config_get_uint32(sn, "delete", &delete_id)) {
-		if (!lv)
-			return SEG_LOG_ERROR("Unknown message in");
-	} else {
-		if (lv)
-			return SEG_LOG_ERROR("Unsupported message format in");
+	} else if (dm_config_get_uint32(sn, "delete", &delete_id))
 		type = DM_THIN_MESSAGE_DELETE;
-	}
+	else
+		return SEG_LOG_ERROR("Unknown message in");
 
-	if (!attach_pool_message(seg, type, lv, delete_id, 1))
+	if (!attach_thin_pool_message(seg, type, lv, delete_id, 1))
 		return_0;
 
 	return 1;
@@ -81,29 +81,31 @@ static int _thin_pool_add_message(struct lv_segment *seg,
 
 static int _thin_pool_text_import(struct lv_segment *seg,
 				  const struct dm_config_node *sn,
-				  struct dm_hash_table *pv_hash __attribute__((unused)))
+				  struct dm_hash_table *pv_hash __attribute__((unused)),
+				  struct dm_hash_table *lv_hash)
 {
 	const char *lv_name;
 	struct logical_volume *pool_data_lv, *pool_metadata_lv;
 	const char *discards_str = NULL;
+	uint32_t zero = 0;
+	uint32_t crop = 0;
 
 	if (!dm_config_get_str(sn, "metadata", &lv_name))
 		return SEG_LOG_ERROR("Metadata must be a string in");
 
-	if (!(pool_metadata_lv = find_lv(seg->lv->vg, lv_name)))
+	if (!(pool_metadata_lv = dm_hash_lookup(lv_hash, lv_name)))
 		return SEG_LOG_ERROR("Unknown metadata %s in", lv_name);
 
 	if (!dm_config_get_str(sn, "pool", &lv_name))
 		return SEG_LOG_ERROR("Pool must be a string in");
 
-	if (!(pool_data_lv = find_lv(seg->lv->vg, lv_name)))
+	if (!(pool_data_lv = dm_hash_lookup(lv_hash, lv_name)))
 		return SEG_LOG_ERROR("Unknown pool %s in", lv_name);
 
-	seg->lv->status |= THIN_POOL;
-	if (!attach_pool_metadata_lv(seg, pool_metadata_lv))
+	if (!attach_pool_data_lv(seg, pool_data_lv))
 		return_0;
 
-	if (!attach_pool_data_lv(seg, pool_data_lv))
+	if (!attach_pool_metadata_lv(seg, pool_metadata_lv))
 		return_0;
 
 	if (!dm_config_get_uint64(sn, "transaction_id", &seg->transaction_id))
@@ -118,25 +120,30 @@ static int _thin_pool_text_import(struct lv_segment *seg,
 
 	if (!discards_str)
 		seg->discards = THIN_DISCARDS_IGNORE;
-	else if (!get_pool_discards(discards_str, &seg->discards))
+	else if (!set_pool_discards(&seg->discards, discards_str))
 		return SEG_LOG_ERROR("Discards option unsupported for");
 
-	if (dm_config_has_node(sn, "low_water_mark") &&
-	    !dm_config_get_uint64(sn, "low_water_mark", &seg->low_water_mark))
-		return SEG_LOG_ERROR("Could not read low_water_mark");
-
 	if ((seg->chunk_size < DM_THIN_MIN_DATA_BLOCK_SIZE) ||
 	    (seg->chunk_size > DM_THIN_MAX_DATA_BLOCK_SIZE))
 		return SEG_LOG_ERROR("Unsupported value %u for chunk_size",
 				     seg->device_id);
 
 	if (dm_config_has_node(sn, "zero_new_blocks") &&
-	    !dm_config_get_uint32(sn, "zero_new_blocks", &seg->zero_new_blocks))
+	    !dm_config_get_uint32(sn, "zero_new_blocks", &zero))
 		return SEG_LOG_ERROR("Could not read zero_new_blocks for");
 
+	seg->zero_new_blocks = (zero) ? THIN_ZERO_YES : THIN_ZERO_NO;
+
+	if (dm_config_has_node(sn, "crop_metadata")) {
+		if (!dm_config_get_uint32(sn, "crop_metadata", &crop))
+			return SEG_LOG_ERROR("Could not read crop_metadata for");
+		seg->crop_metadata = (crop) ? THIN_CROP_METADATA_YES : THIN_CROP_METADATA_NO;
+		seg->lv->status |= LV_CROP_METADATA;
+	}
+
 	/* Read messages */
 	for (; sn; sn = sn->sib)
-		if (!(sn->v) && !_thin_pool_add_message(seg, sn->key, sn->child))
+		if (!(sn->v) && !_thin_pool_add_message(seg, sn->key, sn->child, lv_hash))
 			return_0;
 
 	return 1;
@@ -172,11 +179,16 @@ static int _thin_pool_text_export(const struct lv_segment *seg, struct formatter
 		return 0;
 	}
 
-	if (seg->low_water_mark)
-		outf(f, "low_water_mark = %" PRIu64, seg->low_water_mark);
-
-	if (seg->zero_new_blocks)
+	if (seg->zero_new_blocks == THIN_ZERO_YES)
 		outf(f, "zero_new_blocks = 1");
+	else if (seg->zero_new_blocks != THIN_ZERO_NO) {
+		log_error(INTERNAL_ERROR "Invalid zero new blocks value %d.",
+			  seg->zero_new_blocks);
+		return 0;
+	}
+
+	if (seg->crop_metadata != THIN_CROP_METADATA_UNSELECTED)
+		outf(f, "crop_metadata = %u", (seg->crop_metadata == THIN_CROP_METADATA_YES) ? 1 : 0);
 
 	dm_list_iterate_items(tmsg, &seg->thin_messages) {
 		/* Extra validation */
@@ -221,6 +233,37 @@ static int _thin_pool_text_export(const struct lv_segment *seg, struct formatter
 }
 
 #ifdef DEVMAPPER_SUPPORT
+static int _thin_target_present(struct cmd_context *cmd,
+				const struct lv_segment *seg __attribute__((unused)),
+				unsigned *attributes);
+
+static int _thin_pool_modules_needed(struct dm_pool *mem,
+				     const struct lv_segment *seg __attribute__((unused)),
+				     struct dm_list *modules)
+{
+	if (!str_list_add(mem, modules, _thin_pool_module)) {
+		log_error("String list allocation failed for thin_pool.");
+		return 0;
+	}
+
+	return 1;
+}
+
+static int _thin_modules_needed(struct dm_pool *mem,
+				const struct lv_segment *seg,
+				struct dm_list *modules)
+{
+	if (!_thin_pool_modules_needed(mem, seg, modules))
+		return_0;
+
+	if (!str_list_add(mem, modules, _thin_module)) {
+		log_error("String list allocation failed for thin.");
+		return 0;
+	}
+
+	return 1;
+}
+
 static int _thin_pool_add_target_line(struct dev_manager *dm,
 				      struct dm_pool *mem,
 				      struct cmd_context *cmd,
@@ -231,73 +274,88 @@ static int _thin_pool_add_target_line(struct dev_manager *dm,
 				      uint32_t *pvmove_mirror_count __attribute__((unused)))
 {
 	static int _no_discards = 0;
+	static int _no_error_if_no_space = 0;
 	char *metadata_dlid, *pool_dlid;
 	const struct lv_thin_message *lmsg;
 	const struct logical_volume *origin;
-	struct lvinfo info;
-	uint64_t transaction_id = 0;
 	unsigned attr;
+	uint64_t low_water_mark;
+	int threshold;
 
-	if (!_thin_target_present(cmd, seg, &attr))
+	if (!_thin_target_present(cmd, NULL, &attr))
 		return_0;
 
+	if (!seg->metadata_lv) {
+		log_error(INTERNAL_ERROR "Thin pool is missing metadata device.");
+		return 0;
+	}
+
 	if (!(attr & THIN_FEATURE_BLOCK_SIZE) &&
-	    (seg->chunk_size & (seg->chunk_size - 1))) {
-		log_error("Thin pool target does not support %uKiB chunk size "
-			  "(needs kernel >= 3.6).", seg->chunk_size / 2);
+	    !is_power_of_2(seg->chunk_size)) {
+		log_error("Thin pool target does not support %s chunk size (needs"
+			  " kernel >= 3.6).", display_size(cmd, seg->chunk_size));
 		return 0;
 	}
 
-	if (!laopts->real_pool) {
-		if (!(pool_dlid = build_dm_uuid(mem, seg->lv->lvid.s, "tpool"))) {
-			log_error("Failed to build uuid for thin pool LV %s.", seg->pool_lv->name);
-			return 0;
-		}
-
-		if (!add_linear_area_to_dtree(node, len, seg->lv->vg->extent_size,
-					      cmd->use_linear_target,
-					      seg->lv->vg->name, seg->lv->name) ||
-		    !dm_tree_node_add_target_area(node, NULL, pool_dlid, 0))
-			return_0;
-
-		return 1;
-	}
-
-	if (!(metadata_dlid = build_dm_uuid(mem, seg->metadata_lv->lvid.s, NULL))) {
+	if (!(metadata_dlid = build_dm_uuid(mem, seg->metadata_lv, NULL))) {
 		log_error("Failed to build uuid for metadata LV %s.",
-			  seg->metadata_lv->name);
+			  display_lvname(seg->metadata_lv));
 		return 0;
 	}
 
-	if (!(pool_dlid = build_dm_uuid(mem, seg_lv(seg, 0)->lvid.s, NULL))) {
+	if (!(pool_dlid = build_dm_uuid(mem, seg_lv(seg, 0), NULL))) {
 		log_error("Failed to build uuid for pool LV %s.",
-			  seg_lv(seg, 0)->name);
+			  display_lvname(seg_lv(seg, 0)));
 		return 0;
 	}
 
-	if (!dm_tree_node_add_thin_pool_target(node, len, seg->transaction_id,
-					       metadata_dlid, pool_dlid,
-					       seg->chunk_size, seg->low_water_mark,
-					       seg->zero_new_blocks ? 0 : 1))
+	threshold = find_config_tree_int(seg->lv->vg->cmd,
+					 activation_thin_pool_autoextend_threshold_CFG,
+					 lv_config_profile(seg->lv));
+	if (threshold < 50)
+		threshold = 50;
+	if (threshold < 100)
+		/* Translate to number of free pool blocks to trigger watermark */
+		low_water_mark = len / seg->chunk_size * (100 - threshold) / 100;
+	else
+		low_water_mark = 0;
+
+	if (!dm_tree_node_add_thin_pool_target_v1(node, len,
+						  seg->transaction_id,
+						  metadata_dlid, pool_dlid,
+						  seg->chunk_size, low_water_mark,
+						  (seg->zero_new_blocks == THIN_ZERO_YES) ? 0 : 1,
+						  (seg->crop_metadata == THIN_CROP_METADATA_YES) ? 1 : 0))
 		return_0;
 
 	if (attr & THIN_FEATURE_DISCARDS) {
+		/* Use ignore for discards ignore or non-power-of-2 chunk_size and <1.5 target */
 		/* FIXME: Check whether underlying dev supports discards */
-		if (!dm_tree_node_set_thin_pool_discard(node,
-							seg->discards == THIN_DISCARDS_IGNORE,
-							seg->discards == THIN_DISCARDS_NO_PASSDOWN))
+		if (((!(attr & THIN_FEATURE_DISCARDS_NON_POWER_2) &&
+		      !is_power_of_2(seg->chunk_size)) ||
+		     (seg->discards == THIN_DISCARDS_IGNORE))) {
+			if (!dm_tree_node_set_thin_pool_discard(node, 1, 0))
+				return_0;
+		} else if (!dm_tree_node_set_thin_pool_discard(node, 0,
+							       (seg->discards == THIN_DISCARDS_NO_PASSDOWN)))
 			return_0;
 	} else if (seg->discards != THIN_DISCARDS_IGNORE)
 		log_warn_suppress(_no_discards++, "WARNING: Thin pool target does "
 				  "not support discards (needs kernel >= 3.4).");
 
+	if (attr & THIN_FEATURE_ERROR_IF_NO_SPACE)
+		dm_tree_node_set_thin_pool_error_if_no_space(node, lv_is_error_when_full(seg->lv));
+	else if (lv_is_error_when_full(seg->lv))
+		log_warn_suppress(_no_error_if_no_space++, "WARNING: Thin pool target does "
+				  "not support error if no space (needs version >= 1.10).");
+
 	/*
 	 * Add messages only for activation tree.
 	 * Otherwise avoid checking for existence of suspended origin.
-	 * Also transation_id is checked only when snapshot origin is active.
+	 * Also transaction_id is checked only when snapshot origin is active.
 	 * (This might change later)
 	 */
-	if (!laopts->is_activate)
+	if (!laopts->send_messages)
 		return 1;
 
 	dm_list_iterate_items(lmsg, &seg->thin_messages) {
@@ -305,22 +363,7 @@ static int _thin_pool_add_target_line(struct dev_manager *dm,
 		case DM_THIN_MESSAGE_CREATE_THIN:
 			origin = first_seg(lmsg->u.lv)->origin;
 			/* Check if the origin is suspended */
-			if (origin && lv_info(cmd, origin, 0, &info, 0, 0) &&
-			    info.exists && !info.suspended) {
-				/* Origin is not suspended, but the transaction may have been
-				 * already transfered, so test for transaction_id and
-				 * allow to pass in the message for dmtree processing
-				 * so it will skip all messages later.
-				 */
-				if (!lv_thin_pool_transaction_id(seg->lv, &transaction_id))
-					return_0; /* Thin pool should exist and work */
-				if (transaction_id != seg->transaction_id) {
-					log_error("Can't create snapshot %s as origin %s is not suspended.",
-						  lmsg->u.lv->name, origin->name);
-					return 0;
-				}
-			}
-			log_debug("Thin pool create_%s %s.", (!origin) ? "thin" : "snap", lmsg->u.lv->name);
+			log_debug_activation("Thin pool create_%s %s.", (!origin) ? "thin" : "snap", lmsg->u.lv->name);
 			if (!dm_tree_node_add_thin_pool_message(node,
 								(!origin) ? lmsg->type : DM_THIN_MESSAGE_CREATE_SNAP,
 								first_seg(lmsg->u.lv)->device_id,
@@ -328,7 +371,7 @@ static int _thin_pool_add_target_line(struct dev_manager *dm,
 				return_0;
 			break;
 		case DM_THIN_MESSAGE_DELETE:
-			log_debug("Thin pool delete %u.", lmsg->u.delete_id);
+			log_debug_activation("Thin pool delete %u.", lmsg->u.delete_id);
 			if (!dm_tree_node_add_thin_pool_message(node,
 								lmsg->type,
 								lmsg->u.delete_id, 0))
@@ -342,7 +385,7 @@ static int _thin_pool_add_target_line(struct dev_manager *dm,
 
 	if (!dm_list_empty(&seg->thin_messages)) {
 		/* Messages were passed, modify transaction_id as the last one */
-		log_debug("Thin pool set transaction id %" PRIu64 ".", seg->transaction_id);
+		log_debug_activation("Thin pool set transaction id %" PRIu64 ".", seg->transaction_id);
 		if (!dm_tree_node_add_thin_pool_message(node,
 							DM_THIN_MESSAGE_SET_TRANSACTION_ID,
 							seg->transaction_id - 1,
@@ -354,7 +397,7 @@ static int _thin_pool_add_target_line(struct dev_manager *dm,
 }
 
 static int _thin_pool_target_percent(void **target_state __attribute__((unused)),
-				     percent_t *percent,
+				     dm_percent_t *percent,
 				     struct dm_pool *mem,
 				     struct cmd_context *cmd __attribute__((unused)),
 				     struct lv_segment *seg,
@@ -367,15 +410,17 @@ static int _thin_pool_target_percent(void **target_state __attribute__((unused))
 	if (!dm_get_status_thin_pool(mem, params, &s))
 		return_0;
 
-	/* With 'seg' report metadata percent, otherwice data percent */
-	if (seg) {
-		*percent = make_percent(s->used_metadata_blocks,
-					s->total_metadata_blocks);
+	if (s->fail || s->error)
+		*percent = DM_PERCENT_INVALID;
+	/* With 'seg' report metadata percent, otherwise data percent */
+	else if (seg) {
+		*percent = dm_make_percent(s->used_metadata_blocks,
+					   s->total_metadata_blocks);
 		*total_numerator += s->used_metadata_blocks;
 		*total_denominator += s->total_metadata_blocks;
 	} else {
-		*percent = make_percent(s->used_data_blocks,
-					s->total_data_blocks);
+		*percent = dm_make_percent(s->used_data_blocks,
+					   s->total_data_blocks);
 		*total_numerator += s->used_data_blocks;
 		*total_denominator += s->total_data_blocks;
 	}
@@ -384,18 +429,12 @@ static int _thin_pool_target_percent(void **target_state __attribute__((unused))
 }
 
 #  ifdef DMEVENTD
-static const char *_get_thin_dso_path(struct cmd_context *cmd)
-{
-	return get_monitor_dso_path(cmd, find_config_tree_str(cmd, "dmeventd/thin_library",
-							      DEFAULT_DMEVENTD_THIN_LIB));
-}
-
 /* FIXME Cache this */
-static int _target_registered(struct lv_segment *seg, int *pending)
+static int _target_registered(struct lv_segment *seg, int *pending, int *monitored)
 {
 	return target_registered_with_dmeventd(seg->lv->vg->cmd,
-					       _get_thin_dso_path(seg->lv->vg->cmd),
-					       seg->lv, pending);
+					       seg->segtype->dso,
+					       seg->lv, pending, monitored);
 }
 
 /* FIXME This gets run while suspended and performs banned operations. */
@@ -403,7 +442,7 @@ static int _target_set_events(struct lv_segment *seg, int evmask, int set)
 {
 	/* FIXME Make timeout (10) configurable */
 	return target_register_events(seg->lv->vg->cmd,
-				      _get_thin_dso_path(seg->lv->vg->cmd),
+				      seg->segtype->dso,
 				      seg->lv, evmask, set, 10);
 }
 
@@ -418,25 +457,30 @@ static int _target_unregister_events(struct lv_segment *seg,
 {
 	return _target_set_events(seg, events, 0);
 }
+
 #  endif /* DMEVENTD */
 #endif /* DEVMAPPER_SUPPORT */
 
-static const char *_thin_name(const struct lv_segment *seg)
+static void _thin_display(const struct lv_segment *seg)
 {
-	return seg->segtype->name;
+	log_print("  Device ID\t\t%u", seg->device_id);
+
+	log_print(" ");
 }
 
 static int _thin_text_import(struct lv_segment *seg,
 			     const struct dm_config_node *sn,
-			     struct dm_hash_table *pv_hash __attribute__((unused)))
+			     struct dm_hash_table *pv_hash __attribute__((unused)),
+			     struct dm_hash_table *lv_hash)
 {
 	const char *lv_name;
-	struct logical_volume *pool_lv, *origin = NULL;
+	struct logical_volume *pool_lv, *origin = NULL, *external_lv = NULL, *merge_lv = NULL;
+	struct generic_logical_volume *indirect_origin = NULL;
 
 	if (!dm_config_get_str(sn, "thin_pool", &lv_name))
 		return SEG_LOG_ERROR("Thin pool must be a string in");
 
-	if (!(pool_lv = find_lv(seg->lv->vg, lv_name)))
+	if (!(pool_lv = dm_hash_lookup(lv_hash, lv_name)))
 		return SEG_LOG_ERROR("Unknown thin pool %s in", lv_name);
 
 	if (!dm_config_get_uint64(sn, "transaction_id", &seg->transaction_id))
@@ -446,10 +490,17 @@ static int _thin_text_import(struct lv_segment *seg,
 		if (!dm_config_get_str(sn, "origin", &lv_name))
 			return SEG_LOG_ERROR("Origin must be a string in");
 
-		if (!(origin = find_lv(seg->lv->vg, lv_name)))
+		if (!(origin = dm_hash_lookup(lv_hash, lv_name)))
 			return SEG_LOG_ERROR("Unknown origin %s in", lv_name);
 	}
 
+	if (dm_config_has_node(sn, "merge")) {
+		if (!dm_config_get_str(sn, "merge", &lv_name))
+			return SEG_LOG_ERROR("Merge lv must be a string in");
+		if (!(merge_lv = dm_hash_lookup(lv_hash, lv_name)))
+			return SEG_LOG_ERROR("Unknown merge lv %s in", lv_name);
+	}
+
 	if (!dm_config_get_uint32(sn, "device_id", &seg->device_id))
 		return SEG_LOG_ERROR("Could not read device_id for");
 
@@ -457,7 +508,18 @@ static int _thin_text_import(struct lv_segment *seg,
 		return SEG_LOG_ERROR("Unsupported value %u for device_id",
 				     seg->device_id);
 
-	if (!attach_pool_lv(seg, pool_lv, origin))
+	if (dm_config_has_node(sn, "external_origin")) {
+		if (!dm_config_get_str(sn, "external_origin", &lv_name))
+			return SEG_LOG_ERROR("External origin must be a string in");
+
+		if (!(external_lv = dm_hash_lookup(lv_hash, lv_name)))
+			return SEG_LOG_ERROR("Unknown external origin %s in", lv_name);
+	}
+
+	if (!attach_pool_lv(seg, pool_lv, origin, indirect_origin, merge_lv))
+		return_0;
+
+	if (!attach_thin_external_origin(seg, external_lv))
 		return_0;
 
 	return 1;
@@ -469,9 +531,14 @@ static int _thin_text_export(const struct lv_segment *seg, struct formatter *f)
 	outf(f, "transaction_id = %" PRIu64, seg->transaction_id);
 	outf(f, "device_id = %d", seg->device_id);
 
+	if (seg->external_lv)
+		outf(f, "external_origin = \"%s\"", seg->external_lv->name);
 	if (seg->origin)
 		outf(f, "origin = \"%s\"", seg->origin->name);
 
+	if (seg->merge_lv)
+		outf(f, "merge = \"%s\"", seg->merge_lv->name);
+
 	return 1;
 }
 
@@ -481,27 +548,72 @@ static int _thin_add_target_line(struct dev_manager *dm,
 				 struct cmd_context *cmd __attribute__((unused)),
 				 void **target_state __attribute__((unused)),
 				 struct lv_segment *seg,
-				 const struct lv_activate_opts *laopts __attribute__((unused)),
+				 const struct lv_activate_opts *laopts,
 				 struct dm_tree_node *node, uint64_t len,
 				 uint32_t *pvmove_mirror_count __attribute__((unused)))
 {
-	char *pool_dlid;
+	char *pool_dlid, *external_dlid;
 	uint32_t device_id = seg->device_id;
+	unsigned attr;
 
-	if (!(pool_dlid = build_dm_uuid(mem, seg->pool_lv->lvid.s, "tpool"))) {
+	if (!seg->pool_lv) {
+		log_error(INTERNAL_ERROR "Segment %s has no pool.",
+			  seg->lv->name);
+		return 0;
+	}
+	if (!(pool_dlid = build_dm_uuid(mem, seg->pool_lv, lv_layer(seg->pool_lv)))) {
 		log_error("Failed to build uuid for pool LV %s.",
 			  seg->pool_lv->name);
 		return 0;
 	}
 
+	if (!laopts->no_merging) {
+		if (seg->merge_lv) {
+			log_error(INTERNAL_ERROR "Failed to add merged segment of %s.",
+				  seg->lv->name);
+			return 0;
+		}
+		/*
+		 * merge support for thinp snapshots is implemented by
+		 * simply swapping the thinp device_id of the snapshot
+		 * and origin.
+		 */
+		if (lv_is_merging_origin(seg->lv) && seg_is_thin_volume(find_snapshot(seg->lv)))
+			/* origin, use merging snapshot's device_id */
+			device_id = find_snapshot(seg->lv)->device_id;
+	}
+
 	if (!dm_tree_node_add_thin_target(node, len, pool_dlid, device_id))
 		return_0;
 
+	/* Add external origin LV */
+	if (seg->external_lv) {
+		if (!thin_pool_supports_external_origin(first_seg(seg->pool_lv), seg->external_lv))
+			return_0;
+		if (seg->external_lv->size < seg->lv->size) {
+			/* Validate target supports smaller external origin */
+			if (!_thin_target_present(cmd, NULL, &attr) ||
+			    !(attr & THIN_FEATURE_EXTERNAL_ORIGIN_EXTEND)) {
+				log_error("Thin target does not support smaller size of external origin LV %s.",
+					  seg->external_lv->name);
+				return 0;
+			}
+		}
+		if (!(external_dlid = build_dm_uuid(mem, seg->external_lv,
+						    lv_layer(seg->external_lv)))) {
+			log_error("Failed to build uuid for external origin LV %s.",
+				  seg->external_lv->name);
+			return 0;
+		}
+		if (!dm_tree_node_set_thin_external_origin(node, external_dlid))
+			return_0;
+	}
+
 	return 1;
 }
 
 static int _thin_target_percent(void **target_state __attribute__((unused)),
-				percent_t *percent,
+				dm_percent_t *percent,
 				struct dm_pool *mem,
 				struct cmd_context *cmd __attribute__((unused)),
 				struct lv_segment *seg,
@@ -510,17 +622,32 @@ static int _thin_target_percent(void **target_state __attribute__((unused)),
 				uint64_t *total_denominator)
 {
 	struct dm_status_thin *s;
+	uint64_t csize;
 
 	/* Status for thin device is in sectors */
 	if (!dm_get_status_thin(mem, params, &s))
 		return_0;
 
-	if (seg) {
-		*percent = make_percent(s->mapped_sectors, seg->lv->size);
-		*total_denominator += seg->lv->size;
+	if (s->fail)
+		*percent = DM_PERCENT_INVALID;
+	else if (seg) {
+		/* Pool allocates whole chunk so round-up to nearest one */
+		csize = first_seg(seg->pool_lv)->chunk_size;
+		csize = ((seg->lv->size + csize - 1) / csize) * csize;
+		if (s->mapped_sectors > csize) {
+			log_warn("WARNING: LV %s maps %s while the size is only %s.",
+				 display_lvname(seg->lv),
+				 display_size(cmd, s->mapped_sectors),
+				 display_size(cmd, csize));
+			/* Don't show nonsense numbers like i.e. 1000% full */
+			s->mapped_sectors = csize;
+		}
+
+		*percent = dm_make_percent(s->mapped_sectors, csize);
+		*total_denominator += csize;
 	} else {
 		/* No lv_segment info here */
-		*percent = PERCENT_INVALID;
+		*percent = DM_PERCENT_INVALID;
 		/* FIXME: Using denominator to pass the mapped info upward? */
 		*total_denominator += s->highest_mapped_sector;
 	}
@@ -531,69 +658,96 @@ static int _thin_target_percent(void **target_state __attribute__((unused)),
 }
 
 static int _thin_target_present(struct cmd_context *cmd,
-				const struct lv_segment *seg,
+				const struct lv_segment *seg __attribute__((unused)),
 				unsigned *attributes)
 {
+	/* List of features with their kernel target version */
+	static const struct feature {
+		uint16_t maj;
+		uint16_t min;
+		uint16_t thin_feature;
+		const char feature[24];
+	} _features[] = {
+		{ 1, 1, THIN_FEATURE_DISCARDS, "discards" },
+		{ 1, 1, THIN_FEATURE_EXTERNAL_ORIGIN, "external_origin" },
+		{ 1, 4, THIN_FEATURE_BLOCK_SIZE, "block_size" },
+		{ 1, 5, THIN_FEATURE_DISCARDS_NON_POWER_2, "discards_non_power_2" },
+		{ 1, 10, THIN_FEATURE_METADATA_RESIZE, "metadata_resize" },
+		{ 1, 10, THIN_FEATURE_ERROR_IF_NO_SPACE, "error_if_no_space" },
+		{ 1, 13, THIN_FEATURE_EXTERNAL_ORIGIN_EXTEND, "external_origin_extend" },
+	};
+
+	static const char _lvmconf[] = "global/thin_disabled_features";
 	static int _checked = 0;
 	static int _present = 0;
-	static int _attrs = 0;
+	static unsigned _attrs = 0;
 	uint32_t maj, min, patchlevel;
+	unsigned i;
+	const struct dm_config_node *cn;
+	const struct dm_config_value *cv;
+	const char *str;
+
+	if (!activation())
+		return 0;
 
 	if (!_checked) {
-		_present = target_present(cmd, THIN_MODULE, 1);
-
-		if (!target_version(THIN_MODULE, &maj, &min, &patchlevel)) {
-			log_error("Cannot read " THIN_MODULE " target version.");
-			return 0;
-		}
-
-		if (maj >=1 && min >= 1)
-			_attrs |= THIN_FEATURE_DISCARDS;
-		else
-		/* FIXME Log this as WARNING later only if the user asked for the feature to be used but it's not present */
-			log_debug("Target " THIN_MODULE " does not support discards.");
-
-		if (maj >=1 && min >= 1)
-			_attrs |= THIN_FEATURE_EXTERNAL_ORIGIN;
-		else
-		/* FIXME Log this as WARNING later only if the user asked for the feature to be used but it's not present */
-			log_debug("Target " THIN_MODULE " does not support external origins.");
-
-		if (maj >=1 && min >= 4)
-			_attrs |= THIN_FEATURE_BLOCK_SIZE;
-		else
-		/* FIXME Log this as WARNING later only if the user asked for the feature to be used but it's not present */
-			log_debug("Target " THIN_MODULE " does not support non power of 2 block sizes.");
-
 		_checked = 1;
+
+		if (!(_present = target_present_version(cmd, _thin_pool_module, 1,
+							&maj, &min, &patchlevel)))
+			return 0;
+
+		for (i = 0; i < DM_ARRAY_SIZE(_features); ++i)
+			if ((maj > _features[i].maj) ||
+			    (maj == _features[i].maj && min >= _features[i].min))
+				_attrs |= _features[i].thin_feature;
+			else
+				log_very_verbose("Target %s does not support %s.",
+						 _thin_pool_module,
+						 _features[i].feature);
 	}
 
-	if (attributes)
-		*attributes = _attrs;
+	if (attributes) {
+		if (!_feature_mask) {
+			/* Support runtime lvm.conf changes, N.B. avoid 32 feature */
+			if ((cn = find_config_tree_array(cmd, global_thin_disabled_features_CFG, NULL))) {
+				for (cv = cn->v; cv; cv = cv->next) {
+					if (cv->type != DM_CFG_STRING) {
+						log_warn("WARNING: Ignoring invalid string in config file %s.",
+							  _lvmconf);
+						continue;
+					}
+					str = cv->v.str;
+					if (!*str)
+						continue;
+					for (i = 0; i < DM_ARRAY_SIZE(_features); ++i)
+						if (strcasecmp(str, _features[i].feature) == 0)
+							_feature_mask |= _features[i].thin_feature;
+				}
+			}
+			_feature_mask = ~_feature_mask;
+			for (i = 0; i < DM_ARRAY_SIZE(_features); ++i)
+				if ((_attrs & _features[i].thin_feature) &&
+				    !(_feature_mask & _features[i].thin_feature))
+					log_very_verbose("Target %s %s support disabled by %s",
+							 _thin_pool_module,
+							 _features[i].feature, _lvmconf);
+		}
+		*attributes = _attrs & _feature_mask;
+	}
 
 	return _present;
 }
 #endif
 
-static int _thin_modules_needed(struct dm_pool *mem,
-				const struct lv_segment *seg __attribute__((unused)),
-				struct dm_list *modules)
-{
-	if (!str_list_add(mem, modules, THIN_MODULE)) {
-		log_error("thin string list allocation failed");
-		return 0;
-	}
-
-	return 1;
-}
-
 static void _thin_destroy(struct segment_type *segtype)
 {
-	dm_free(segtype);
+	free((void *) segtype->dso);
+	free(segtype);
 }
 
-static struct segtype_handler _thin_pool_ops = {
-	.name = _thin_pool_name,
+static const struct segtype_handler _thin_pool_ops = {
+	.display = _thin_pool_display,
 	.text_import = _thin_pool_text_import,
 	.text_import_area_count = _thin_pool_text_import_area_count,
 	.text_export = _thin_pool_text_export,
@@ -601,26 +755,26 @@ static struct segtype_handler _thin_pool_ops = {
 	.add_target_line = _thin_pool_add_target_line,
 	.target_percent = _thin_pool_target_percent,
 	.target_present = _thin_target_present,
+	.modules_needed = _thin_pool_modules_needed,
 #  ifdef DMEVENTD
 	.target_monitored = _target_registered,
 	.target_monitor_events = _target_register_events,
 	.target_unmonitor_events = _target_unregister_events,
 #  endif /* DMEVENTD */
 #endif
-	.modules_needed = _thin_modules_needed,
 	.destroy = _thin_destroy,
 };
 
-static struct segtype_handler _thin_ops = {
-	.name = _thin_name,
+static const struct segtype_handler _thin_ops = {
+	.display = _thin_display,
 	.text_import = _thin_text_import,
 	.text_export = _thin_text_export,
 #ifdef DEVMAPPER_SUPPORT
 	.add_target_line = _thin_add_target_line,
 	.target_percent = _thin_target_percent,
 	.target_present = _thin_target_present,
-#endif
 	.modules_needed = _thin_modules_needed,
+#endif
 	.destroy = _thin_destroy,
 };
 
@@ -632,35 +786,38 @@ int init_multiple_segtypes(struct cmd_context *cmd, struct segtype_library *segl
 #endif
 {
 	static const struct {
-		struct segtype_handler *ops;
-		const char name[16];
+		const struct segtype_handler *ops;
+		const char name[12];
 		uint32_t flags;
-	} reg_segtypes[] = {
-		{ &_thin_pool_ops, "thin-pool", SEG_THIN_POOL },
+	} _reg_segtypes[] = {
+		{ &_thin_pool_ops, "thin-pool", SEG_THIN_POOL | SEG_CANNOT_BE_ZEROED |
+		SEG_ONLY_EXCLUSIVE | SEG_CAN_ERROR_WHEN_FULL },
 		/* FIXME Maybe use SEG_THIN_VOLUME instead of SEG_VIRTUAL */
-		{ &_thin_ops, "thin", SEG_THIN_VOLUME | SEG_VIRTUAL }
+		{ &_thin_ops, "thin", SEG_THIN_VOLUME | SEG_VIRTUAL | SEG_ONLY_EXCLUSIVE }
 	};
 
 	struct segment_type *segtype;
 	unsigned i;
 
-	for (i = 0; i < sizeof(reg_segtypes)/sizeof(reg_segtypes[0]); ++i) {
-		segtype = dm_zalloc(sizeof(*segtype));
+	for (i = 0; i < DM_ARRAY_SIZE(_reg_segtypes); ++i) {
+		segtype = zalloc(sizeof(*segtype));
 
 		if (!segtype) {
 			log_error("Failed to allocate memory for %s segtype",
-				  reg_segtypes[i].name);
+				  _reg_segtypes[i].name);
 			return 0;
 		}
 
-		segtype->ops = reg_segtypes[i].ops;
-		segtype->name = reg_segtypes[i].name;
-		segtype->flags = reg_segtypes[i].flags;
+		segtype->ops = _reg_segtypes[i].ops;
+		segtype->name = _reg_segtypes[i].name;
+		segtype->flags = _reg_segtypes[i].flags;
 
 #ifdef DEVMAPPER_SUPPORT
 #  ifdef DMEVENTD
-		if ((reg_segtypes[i].flags & SEG_THIN_POOL) &&
-		    _get_thin_dso_path(cmd))
+		segtype->dso = get_monitor_dso_path(cmd, dmeventd_thin_library_CFG);
+
+		if ((_reg_segtypes[i].flags & SEG_THIN_POOL) &&
+		    segtype->dso)
 			segtype->flags |= SEG_MONITORED;
 #  endif /* DMEVENTD */
 #endif
@@ -671,5 +828,9 @@ int init_multiple_segtypes(struct cmd_context *cmd, struct segtype_library *segl
 		log_very_verbose("Initialised segtype: %s", segtype->name);
 	}
 
+
+	/* Reset mask for recalc */
+	_feature_mask = 0;
+
 	return 1;
 }
